#!/usr/bin/env python

import os
import platform
import cffi

try:
    print("OpenSSL Path: %s" % os.environ["OPENSSL_DIR"])
    openssl_dir = os.environ["OPENSSL_DIR"]
except:
    if (platform.system() == "Darwin"
        and os.path.isdir('/usr/local/opt/openssl/include')):
        openssl_dir = '/usr/local/opt/openssl/include'
    else:
        print("Using default openssl location. Set OPENSSL_DIR env variable to change it.")
        openssl_dir = '../openssl'


# Determine the include and src directory
from os.path import join
csrc = "include" # abspath(join(dirname(__file__),"../src"))

## Asume we are running on a posix system
# LINUX: libraries=["crypto"], extra_compile_args=['-Wno-deprecated-declarations']
link_args = []
libraries=["crypto"]
extra_compile_args=['-Wno-deprecated-declarations']
include_dirs=[csrc, openssl_dir]
library_dirs=[]

cfiles = [r"bp_fp12.c",  r"bp_fp6.c", r"bp_g2.c", 
          r"bp_group.c",  r"bp_map.c", r"bp_fp2.c",   
          r"bp_g1.c", r"bp_g2_mult.c", r"bp_gt.c"]
print("Path: %s" % csrc)
csources = list(map(lambda x: r"bplib/src/"+x,cfiles))

from petlib.compile import _FFI as petlib_ffibuilder
ffibuilder = cffi.FFI()
ffibuilder.include(petlib_ffibuilder)

ffibuilder.set_source("bplib._bplib","""
#include <openssl/crypto.h>
#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/ec.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>
#include <openssl/ecdsa.h>

#include <openssl/ossl_typ.h>

#include <bp.h>

    """, libraries=libraries, 
    extra_compile_args=extra_compile_args, 
    include_dirs=include_dirs,
    library_dirs=library_dirs, 
    extra_link_args=link_args,
    sources=csources)


ffibuilder.cdef("""

// BP Functions
typedef ... BP_GROUP;
typedef ... G1_ELEM;
typedef ... G2_ELEM;
typedef ... GT_ELEM;

#define NID_fp254bnb 1

BP_GROUP *BP_GROUP_new(void);
BP_GROUP *BP_GROUP_new_by_curve_name(int nid);
void BP_GROUP_clear_free(BP_GROUP *group);
BP_GROUP *BP_GROUP_dup(const BP_GROUP *a);
const EC_GROUP *BP_GROUP_get_group_G1(BP_GROUP *group);

int BP_GROUP_get_order(const BP_GROUP *group, BIGNUM *order, BN_CTX *ctx);
int BP_GROUP_get_generator_G1(const BP_GROUP *group, G1_ELEM *g);
int BP_GROUP_precompute_mult_G1(BP_GROUP *group, BN_CTX *ctx);
int BP_GROUP_get_generator_G2(const BP_GROUP *group, G2_ELEM *g);
int BP_GROUP_precompute_mult_G2(BP_GROUP *group, BN_CTX *ctx);
int BP_GROUP_get_curve(const BP_GROUP *group, BIGNUM *p, BIGNUM *a,
                       BIGNUM *b, BN_CTX *ctx);
G1_ELEM *G1_ELEM_new(const BP_GROUP *group);
void G1_ELEM_free(G1_ELEM *point);
void G1_ELEM_clear_free(G1_ELEM *point);
int G1_ELEM_copy(G1_ELEM *dst, const G1_ELEM *src);
G1_ELEM *G1_ELEM_dup(const G1_ELEM *src, const BP_GROUP *group);
int G1_ELEM_set_to_infinity(const BP_GROUP *group, G1_ELEM *point);
int G1_ELEM_set_Jprojective_coordinates(const BP_GROUP *group,
                                        G1_ELEM *point, const BIGNUM *x,
                                        const BIGNUM *y,
                                        const BIGNUM *z, BN_CTX *ctx);
int G1_ELEM_get_Jprojective_coordinates(const BP_GROUP *group,
                                        const G1_ELEM *point, BIGNUM *x,
                                        BIGNUM *y, BIGNUM *z,
                                        BN_CTX *ctx);
int G1_ELEM_set_affine_coordinates(const BP_GROUP *group, G1_ELEM *point,
                                   const BIGNUM *x, const BIGNUM *y,
                                   BN_CTX *ctx);
int G1_ELEM_get_affine_coordinates(const BP_GROUP *group,
                                   const G1_ELEM *point, BIGNUM *x,
                                   BIGNUM *y, BN_CTX *ctx);
int G1_ELEM_set_compressed_coordinates(const BP_GROUP *group,
                                       G1_ELEM *point, const BIGNUM *x,
                                       int y_bit, BN_CTX *ctx);
size_t G1_ELEM_point2oct(const BP_GROUP *group, const G1_ELEM *point,
                         point_conversion_form_t form, unsigned char *buf,
                         size_t len, BN_CTX *ctx);
int G1_ELEM_oct2point(const BP_GROUP *group, const G1_ELEM *point,
                      const unsigned char *buf, size_t len, BN_CTX *ctx);
/********************************************************************/
/*              Functions for arithmetic in group G1                */
/********************************************************************/
int G1_ELEM_add(const BP_GROUP *group, G1_ELEM *r, const G1_ELEM *a,
                const G1_ELEM *b, BN_CTX *ctx);
int G1_ELEM_dbl(const BP_GROUP *group, G1_ELEM *r, const G1_ELEM *a,
                BN_CTX *ctx);
int G1_ELEM_invert(const BP_GROUP *group, G1_ELEM *a, BN_CTX *ctx);
int G1_ELEM_is_at_infinity(const BP_GROUP *group, const G1_ELEM *point);
int G1_ELEM_is_on_curve(const BP_GROUP *group, const G1_ELEM *point,
                        BN_CTX *ctx);
int G1_ELEM_cmp(const BP_GROUP *group, const G1_ELEM *point,
                const G1_ELEM *b, BN_CTX *ctx);
int G1_ELEM_make_affine(const BP_GROUP *group, G1_ELEM *point, BN_CTX *ctx);
int G1_ELEMs_make_affine(const BP_GROUP *group, size_t num,
                         G1_ELEM *points[], BN_CTX *ctx);
int G1_ELEM_mul(const BP_GROUP *group, G1_ELEM *r, const BIGNUM *g_scalar,
                const G1_ELEM *point, const BIGNUM *p_scalar,
                BN_CTX *ctx);
int G1_ELEMs_mul(const BP_GROUP *group, G1_ELEM *r, const BIGNUM *scalar,
                 size_t num, const G1_ELEM *points[],
                 const BIGNUM *scalars[], BN_CTX *ctx);
/********************************************************************/
/*              Functions for managing G2 elements                  */
/********************************************************************/
G2_ELEM *G2_ELEM_new(const BP_GROUP *group);
void G2_ELEM_free(G2_ELEM *point);
void G2_ELEM_clear_free(G2_ELEM *point);
int G2_ELEM_copy(G2_ELEM *dst, const G2_ELEM *src);
G2_ELEM *G2_ELEM_dup(const G2_ELEM *src, const BP_GROUP *group);
/********************************************************************/
/*              Functions for arithmetic in group G2                */
/********************************************************************/
int G2_ELEM_set_to_infinity(const BP_GROUP *group, G2_ELEM *point);
int G2_ELEM_set_Jprojective_coordinates(const BP_GROUP *group,
                                        G2_ELEM *point, const BIGNUM *x[2],
                                        const BIGNUM *y[2],
                                        const BIGNUM *z[2], BN_CTX *ctx);
int G2_ELEM_get_Jprojective_coordinates(const BP_GROUP *group,
                                        const G2_ELEM *point, BIGNUM *x[2],
                                        BIGNUM *y[2], BIGNUM *z[2],
                                        BN_CTX *ctx);
int G2_ELEM_set_affine_coordinates(const BP_GROUP *group, G2_ELEM *point,
                                   const BIGNUM *x[2], const BIGNUM *y[2],
                                   BN_CTX *ctx);
int G2_ELEM_get_affine_coordinates(const BP_GROUP *group,
                                   const G2_ELEM *point, BIGNUM *x[2], BIGNUM *y[2],
                                   BN_CTX *ctx);
size_t G2_ELEM_point2oct(const BP_GROUP *group, const G2_ELEM *point,
                         point_conversion_form_t form, unsigned char *buf,
                         size_t len, BN_CTX *ctx);
int G2_ELEM_oct2point(const BP_GROUP *group, G2_ELEM *point,
                      const unsigned char *buf, size_t len, BN_CTX *ctx);
/********************************************************************/
/*              Functions for arithmetic in group G2                */
/********************************************************************/
int G2_ELEM_add(const BP_GROUP *group, G2_ELEM *r, const G2_ELEM *a,
                const G2_ELEM *b, BN_CTX *ctx);
int G2_ELEM_dbl(const BP_GROUP *group, G2_ELEM *r, const G2_ELEM *a,
                BN_CTX *ctx);
int G2_ELEM_invert(const BP_GROUP *group, G2_ELEM *point, BN_CTX *ctx);
int G2_ELEM_is_at_infinity(const BP_GROUP *group, const G2_ELEM *point);
int G2_ELEM_is_on_curve(const BP_GROUP *group, const G2_ELEM *point,
                        BN_CTX *ctx);
int G2_ELEM_cmp(const BP_GROUP *group, const G2_ELEM *point,
                const G2_ELEM *b, BN_CTX *ctx);
int G2_ELEM_make_affine(const BP_GROUP *group, G2_ELEM *point, BN_CTX *ctx);
int G2_ELEMs_make_affine(const BP_GROUP *group, size_t num,
                         G2_ELEM *points[], BN_CTX *ctx);
int G2_ELEM_mul(const BP_GROUP *group, G2_ELEM *r, const BIGNUM *g_scalar,
                const G2_ELEM *point, const BIGNUM *p_scalar,
                BN_CTX *ctx);
int G2_ELEMs_mul(const BP_GROUP *group, G2_ELEM *r, const BIGNUM *scalar,
                 size_t num, const G2_ELEM *points[],
                 const BIGNUM *scalars[], BN_CTX *ctx);
/********************************************************************/
/*              Functions for managing GT elements                  */
/********************************************************************/
GT_ELEM *GT_ELEM_new(const BP_GROUP *group);
void GT_ELEM_free(GT_ELEM *elem);
void GT_clear_free(GT_ELEM *a);
int GT_ELEM_copy(GT_ELEM *dst, const GT_ELEM *src);
GT_ELEM *GT_ELEM_dup(const GT_ELEM *src, const BP_GROUP *group);
int GT_ELEM_zero(GT_ELEM *a);
int GT_ELEM_is_zero(GT_ELEM *a);
int GT_ELEM_set_to_unity(const BP_GROUP *group, GT_ELEM *a);
int GT_ELEM_is_unity(const BP_GROUP *group, const GT_ELEM *a);
size_t GT_ELEM_elem2oct(const BP_GROUP *group, const GT_ELEM *a,
                         unsigned char *buf, size_t len, BN_CTX *ctx);
int GT_ELEM_oct2elem(const BP_GROUP *group, GT_ELEM *a,
                      const unsigned char *buf, size_t len, BN_CTX *ctx);
int GT_ELEM_add(const BP_GROUP *group, GT_ELEM *r, const GT_ELEM *a,
                const GT_ELEM *b, BN_CTX *ctx);
int GT_ELEM_sub(const BP_GROUP *group, GT_ELEM *r, const GT_ELEM *a,
                const GT_ELEM *b, BN_CTX *ctx);
int GT_ELEM_sqr(const BP_GROUP *group, GT_ELEM *r, const GT_ELEM *a,
                BN_CTX *ctx);
int GT_ELEM_mul(const BP_GROUP *group, GT_ELEM *r, GT_ELEM *a, GT_ELEM *b, BN_CTX *ctx);
int GT_ELEM_inv(const BP_GROUP *group, GT_ELEM *r, GT_ELEM *a, BN_CTX *ctx);
int GT_ELEM_cmp(const GT_ELEM *a, const GT_ELEM *b);
int GT_ELEM_exp(const BP_GROUP *group, GT_ELEM *r, const GT_ELEM *a, const BIGNUM *b,
                BN_CTX *ctx);
int GT_ELEM_pairing(const BP_GROUP *group, GT_ELEM *r, const G1_ELEM *p,
                    const G2_ELEM *q, BN_CTX *ctx);
int GT_ELEMs_pairing(const BP_GROUP *group, GT_ELEM *r, size_t num,
                     const G1_ELEM *p[], const G2_ELEM *q[], BN_CTX *ctx);


""")

if __name__ == "__main__":
    print("Compiling bp ...")
    ffibuilder.compile(verbose=True)



from .bp import BpGroup, G1Elem, G2Elem, GTElem

try:
    from petlib.pack import encode, decode, register_coders
    import msgpack
except:
    print("Avoid imports to compile docs")

def g_enc(obj):
    return msgpack.packb( obj.nid )

def g_dec(data):
    nid = msgpack.unpackb(data)
    return BpGroup(nid)

def pt_enc(obj):
    return msgpack.packb( (obj.group.nid, obj.export() ) )

def pt_decoder(xtype):
    def dec(data):
        (nid, bts) = msgpack.unpackb(data)
        G = BpGroup(nid)
        pt = xtype.from_bytes(bts, G) 
        return pt
    return dec

try:
    register_coders(BpGroup, 100, g_enc, g_dec)
    register_coders(G1Elem, 101, pt_enc, pt_decoder(G1Elem))
    register_coders(G2Elem, 102, pt_enc, pt_decoder(G2Elem))
    register_coders(GTElem, 103, pt_enc, pt_decoder(GTElem))
except:
    print("Avoid imports to compile docs")


def test_pack_trivial():
    G = BpGroup()

    # Key Generation
    private = G.order().random()
    pub = private * G.gen2() # The public key

    # Signature
    message = b"Hello World"
    sig = private * G.hashG1(message)
   
    # Verification
    gt = G.pair(sig, G.gen2())
    assert gt == G.pair(G.hashG1(message), pub)

    assert pub == pt_decoder(G2Elem)(pt_enc(pub))
    assert sig == pt_decoder(G1Elem)(pt_enc(sig))
    assert gt == pt_decoder(GTElem)(pt_enc(gt))

def test_pack_full():
    G = BpGroup()

    # Key Generation
    private = G.order().random()
    pub = private * G.gen2() # The public key

    # Signature
    message = b"Hello World"
    sig = private * G.hashG1(message)
   
    # Verification
    gt = G.pair(sig, G.gen2())

    struct = [G, pub, sig, gt]
    data = encode( struct )
    assert struct == decode(data)


""" An implementation of the Pointcheval-Sanders Short Randomizable Signatures 
scheme, to demononstrate the capabilities of the bp module. For full details
of the scheme see:

David Pointcheval, Olivier Sanders: Short Randomizable Signatures. CT-RSA 2016: 111-126

"""

from bplib.bp import BpGroup

def setup():
	G = BpGroup()
	g1, g2 = G.gen1(), G.gen2()
	e, o = G.pair, G.order()
	return (G, o, g1, g2, e)

def keygen(params):
	(G, o, g1, g2, e) = params
	(x, y) = o.random(), o.random()
	sk = (x, y)
	pk = (g2, x*g2, y*g2)
	return (sk, pk)

def sign(params, sk, m):
	(G, o, g1, g2, e) = params
	(x, y) = sk
	h = (o.random()) * g1
	sig = (x+y*m) * h
	return (h, sig)

def verify(params, pk, m, sig):
	(G, o, g1, g2, e) = params
	(g, X, Y) = pk
	sig1 , sig2 = sig
	return not sig1.isinf() and e(sig1, X + m * Y) == e(sig2, g)

def randomize(params, sig):
	(G, o, g1, g2, e) = params
	sig1 , sig2 = sig
	t = o.random()
	return ( t*sig1 , t*sig2 )

# ---------- TESTS -------------

def test_setup():
	setup()

def test_keygen():
	params = setup()
	sk, pk = keygen(params)

def test_sign():
	params = setup()
	sk, pk = keygen(params)

	from petlib.bn import Bn
	from hashlib import sha256

	m = Bn.from_binary(sha256("Hello World!").digest())
	sign(params, sk,  m)

def test_verify():
	params = setup()
	sk, pk = keygen(params)

	from petlib.bn import Bn
	from hashlib import sha256

	m = Bn.from_binary(sha256("Hello World!").digest())
	signature = sign(params, sk,  m)

	assert verify(params, pk, m, signature)

	m2 = Bn.from_binary(sha256("Other Hello World!").digest())
	assert not verify(params, pk, m2, signature)

def test_randomize():
	params = setup()
	sk, pk = keygen(params)

	from petlib.bn import Bn
	from hashlib import sha256

	m = Bn.from_binary(sha256("Hello World!").digest())
	signature = sign(params, sk,  m)

	signature = randomize(params, signature)

	assert verify(params, pk, m, signature)

	m2 = Bn.from_binary(sha256("Other Hello World!").digest())
	assert not verify(params, pk, m2, signature)




try:
    from . import bp
    from . import pack
except:
    pass

__doc__ = """The ``bplib`` is a library implementing support for computations on groups supporting 
bilinear pairings, as used in modern cryptography. 

It is based on the OpenPairing library by 
Diego Aranha (https://github.com/dfaranha/OpenPairing), which is itself based on, and compatible 
with, OpenSSL math functions (``bn`` and ``ec``). The ``bplib`` is compatible with ``petlib`` types
including ``petlib.bn`` and the group G1 is a ``petlib.ec`` EC group. Along with ``petlib``, 
they provide easy to use 
support for maths and ciphers used in modern Privacy Enhancing Technologies."

A set of bilinear EC groups is defined as:

    >>> G = bp.BpGroup()

Such a BpGroup describes 3 groups G1, G2 and GT such that pair(G1,G2)->GT. Generators 
for the groups G1 and G2 are denoted by:

    >>> g1, g2 = G.gen1(), G.gen2()

The special ``pair`` operation computes to pairing into GT:

    >>> gt = G.pair(g1, g2)

Operations are defined on all elements of G1, G2 or GT in a natural additive infix notation for G1 and G2, and a multiplicative notation for GT:
    
    >>> gt6 = gt**6

As expected the ``pair`` operations is additive:

    >>> G.pair(g1, 6*g2) == gt6
    True
    >>> G.pair(6*g1, g2) == gt6
    True
    >>> G.pair(2*g1, 3*g2) == gt6
    True

"""

VERSION = "0.0.63"



""" A Skeleton implementation of the MP3 presence protocol:

MP3: A More Efficient Private Presence Protocol
Rahul Parhi, Michael Schliep, Nicholas Hopper
(Submitted on 10 Sep 2016)
https://arxiv.org/abs/1609.02987

"""

from bplib.bp import BpGroup

def params():
	G = BpGroup()
	Zp = G.order()
	return (G, Zp)

def managerkey(params):
	G, Zp = params()

	Ga = Zp.random() * G.gen1()
	Ha = Zp.random() * G.gen2()
	gamma_a = Zp.random()

	managerkey_a = (Ga, Ha, gamma_a)

	x_af = Zp.random()

	ex = (gamma_a.mod_add(x_af,Zp)).mod_inverse(Zp)

	A_af = x_af.mod_mul(ex,Zp) * Ga
	B_af = ex * Ha

	decryptionkey_a = (x_af, A_af, B_af)

	return managerkey_a, decryptionkey_a

def register(params, managerkey_a):
	G, Zp = params()
	(Ga, Ha, gamma_a) = managerkey_a

	kappa = Zp.random()

	Cj_a1 = kappa*(gamma_a * Ga)
	Cj_a2 = kappa*Ha

	Kj_a = G.pair(Ga, Ha) ** kappa

	record_a = (Cj_a1, Cj_a2)
	key_a = Kj_a
	return (record_a, key_a)	

def lookup(params, decryptionkey_a, record_a):
	G, Zp = params()

	(x_af, A_af, B_af) = decryptionkey_a
	(Cj_a1, Cj_a2) = record_a

	K = G.pair(Cj_a1, B_af).mul(G.pair(A_af, Cj_a2))
	return K

def test_all():

	G, Zp = params()

	# Derive friend key
	managerkey_a, decryptionkey_a = managerkey(params)

	# Registration
	record_a, key_a = register(params, managerkey_a)

	# Lookup
	K = lookup(params, decryptionkey_a, record_a)

	assert key_a.eq( K )


""" A bilinear pairing library based on Diego Aranha's bp contribution to OpenSSL. 
As an example here is an implementation of the BLS signature scheme (Boneh, Lynn, Shacham, 2001):
    >>> G = BpGroup()
    >>>
    >>> # Key Generation
    >>> private = G.order().random()
    >>> pub = private * G.gen2() # The public key
    >>>
    >>> # Signature
    >>> message = b"Hello World"
    >>> sig = private * G.hashG1(message)
    >>>
    >>> # Verification
    >>> G.pair(sig, G.gen2()) == G.pair(G.hashG1(message), pub)
    True
"""


from copy import copy
from binascii import hexlify
from hashlib import sha512

try:
    from petlib.bn import Bn, force_Bn
except:
    print("Do not load petlib to build docs.")

    from functools import wraps

    def force_Bn(n):
        """A decorator that coerces the nth input to be a Big Number"""

        def convert_nth(f):
            # pylint: disable=star-args
            @wraps(f)
            def new_f(*args, **kwargs):
                new_args = args
                return f(*new_args, **kwargs)
            return new_f
        return convert_nth


try:
    from .bindings import _FFI, _C

    from petlib.bindings import _C as petlibc
    import pytest

    from builtins import int        # pylint: disable=redefined-builtin
    from builtins import object     # pylint: disable=redefined-builtin

    POINT_CONVERSION_COMPRESSED = petlibc.POINT_CONVERSION_COMPRESSED
    POINT_CONVERSION_UNCOMPRESSED = petlibc.POINT_CONVERSION_UNCOMPRESSED

except:                             # pylint: disable=bare-except
    POINT_CONVERSION_COMPRESSED = 0
    POINT_CONVERSION_UNCOMPRESSED = 1
    print('Cannot mock for docs')

try:
    from future.utils import python_2_unicode_compatible
except Exception as e:              # pylint: disable=broad-except
    # An identity decorator
    python_2_unicode_compatible = lambda x: x

def _check(return_val):
        """Checks the return code of the C calls"""
        if __debug__:
            if isinstance(return_val, int) and return_val == 1:
                return
            if isinstance(return_val, bool) and return_val == True:
                return

        if return_val == 1 or return_val == True:
            return

        raise Exception("EC exception: %s" % return_val) 


NID_fp254bnb = 1

class BpGroup(object):
    """ A class representing all groups involved in the bilinear pairing: G1, G2, and GT. """
    
    bpq = None
    try:
        _C = _C
    except:
        _C = None
        print("Mock for docs.")

    def __init__(self, nid=NID_fp254bnb, optimize_mult=True):
        """Build an BP group from the Open SSL nid."""

        self.nid = nid

        self.bpg = _C.BP_GROUP_new_by_curve_name(nid)
        self.optimize_mult = optimize_mult

        if optimize_mult:
            _check(_C.BP_GROUP_precompute_mult_G1(self.bpg, _FFI.NULL))
            _check(_C.BP_GROUP_precompute_mult_G2(self.bpg, _FFI.NULL))

        self.g1 = G1Elem(self)
        _check( _C.BP_GROUP_get_generator_G1(self.bpg, self.g1.elem) )

        self.g2 = G2Elem(self)
        _check( _C.BP_GROUP_get_generator_G2(self.bpg, self.g2.elem) )

        self.ord = None
        self.inf = None

        self.p = None

        self.math = _C

    def order(self):
        """Returns the order of the group as a Big Number.
        
        Example:
            >>> G = BpGroup()
            >>> print(G.order())
            16798108731015832284940804142231733909759579603404752749028378864165570215949
        """

        if self.ord is None:
            o = Bn()
            _check( _C.BP_GROUP_get_order(self.bpg, o.bn, _FFI.NULL) )
            self.ord = o

        return self.ord

    def gen1(self):
        """ Returns the generator for G1. """
        return self.g1

    def hashG1(self, sbin):
        """ Hashes a byte string into a point of G1. 
        
        Example:
            >>> G = BpGroup()
            >>> g1 = G.gen1()
            >>> g1p = G.hashG1(b"Hello")
            >>> x = g1 + g1p
        """

        if self.p is None:
            p = Bn()
            a, b = Bn(), Bn()
            _check( _C.BP_GROUP_get_curve(self.bpg, p.bn, a.bn, b.bn, _FFI.NULL) )
            self.p = p

        pt = G1Elem(self)
        xhash = sbin
        y = 1    
        ret = 0

        while ret == 0:
            xhash = sha512(xhash).digest()
            x = Bn.from_binary(xhash) % self.p
            ret = _C.G1_ELEM_set_compressed_coordinates(self.bpg, pt.elem, x.bn, y, _FFI.NULL)

        _check( _C.G1_ELEM_is_on_curve(self.bpg, pt.elem, _FFI.NULL) )
        return pt


    def gen2(self):
        """ Returns the generator for G2. """
        return self.g2

    def pair(self, g1, g2):
        """ The pairing operation e(G1, G2) -> GT. 
        
            Example:
                >>> G = BpGroup()
                >>> g1, g2 = G.gen1(), G.gen2()
                >>> gt = G.pair(g1, g2)
                >>> gt6 = G.pair(g1.mul(2), g2.mul(3))
                >>> gt.exp(6).eq( gt6 )
                True
                >>> gt**6 == G.pair(2*g1, 3*g2)
                True
        """

        gt = GTElem(self)
        _check( _C.GT_ELEM_pairing(self.bpg, gt.elem, g1.elem, g2.elem, _FFI.NULL) )
        return gt


    def __copy__(self):
        return BpGroup(self.nid, self.optimize_mult)

    def __eq__(self, G):
        return G.nid == self.nid

    def __del__(self):
        """ Clears the Group object """
        if self.bpq is not None:
            self._C.BP_GROUP_clear_free(self.bpg)


class Ops(object):
    """ A class to implement infix operations. """

    try:
        _C = _C
    except:
        _C = None
        print("Mock for docs.")

    def __eq__(self, other):
        try:
            return self.eq(other)
        except Exception as e:
            raise Exception("Equality Operation Failed: %s" % e.value())

    def __hash__(self):
        return self.export().__hash__()
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        return self.add(other)

    def __sub__(self, other):
        try:
            return self.sub(other)
        except:
            return self.add(other.neg())

    def __mul__(self, other):
        return self.mul(other)

    def __rmul__(self, other):
        return self.mul(other)

    def __pow__(self, other, modulo=None):
        return self.exp(other)

    def __neg__(self):
        return self.neg()
        
class G1Elem(Ops):

    @staticmethod
    def inf(group):
        """ Returns the element at infinity for G1 """
        zero = G1Elem(group)
        _check( _C.G1_ELEM_set_to_infinity(group.bpg, zero.elem) )
        return zero

    def __init__(self, group):
        """ Returns and element of G1 """
        self.group = group
        self.elem = _C.G1_ELEM_new(group.bpg);

    def __copy__(self):
        """ Copy the G1 point. """
        newpt = G1Elem(self.group)
        _check( _C.G1_ELEM_copy(newpt.elem, self.elem) )
        return newpt

    def add(self, other):
        """ Returns the sum of two points. """
        newpt = G1Elem(self.group)
        _check( _C.G1_ELEM_add(self.group.bpg, newpt.elem, self.elem, other.elem, _FFI.NULL) )
        return newpt   

    def double(self):
        """ Returns the double of the G1 point. """
        newpt = G1Elem(self.group)
        _check( _C.G1_ELEM_dbl(self.group.bpg, newpt.elem, self.elem, _FFI.NULL) )
        return newpt

    def neg(self):
        """ Returns the inverse point. 
        
            Example:
                >>> G = BpGroup()
                >>> g1 = G.gen1()
                >>> g1.add(g1.neg()).isinf()
                True
                >>> g1 - g1 == G1Elem.inf(G)
                True
        """
        newpt = self.__copy__()
        _check( _C.G1_ELEM_invert(self.group.bpg, newpt.elem, _FFI.NULL))
        return newpt

    def eq(self, other):
        """ Returns True if points are equal.
        
            Example:
                >>> G = BpGroup()
                >>> g1 = G.gen1()
                >>> g1.add(g1).eq(g1.double())
                True
                >>> g1.eq(g1.double())
                False
                >>> g1+g1 == 2*g1
                True
                >>> g1+g1 == Bn(2)*g1
                True
        """
        resp = _C.G1_ELEM_cmp(self.group.bpg, self.elem, other.elem, _FFI.NULL)
        return (int(resp) == 0)

    def isinf(self):
        """ Returns True if the element is infinity. """
        return int(_C.G1_ELEM_is_at_infinity(self.group.bpg, self.elem)) == 1

    @force_Bn(1)
    def mul(self, scalar):
        """ Multiplies the point with a scalar. 
        
            Example:
                >>> g1 = BpGroup().gen1()
                >>> g1.mul(2).eq(g1.double())
                True
        """
        newpt = G1Elem(self.group)
        _check( _C.G1_ELEM_mul(self.group.bpg, newpt.elem, _FFI.NULL, self.elem, scalar.bn, _FFI.NULL) )
        return newpt

    def export(self, form=POINT_CONVERSION_COMPRESSED):
        """ Export a point to a byte representation. """
        size = int(_C.G1_ELEM_point2oct(self.group.bpg, self.elem, form, _FFI.NULL, 0, _FFI.NULL))

        out = _FFI.new("unsigned char[]", size)
        _C.G1_ELEM_point2oct(self.group.bpg, self.elem, form, out, size, _FFI.NULL)
        ret = bytes(_FFI.buffer(out)[:size])

        return ret

    @staticmethod
    def from_bytes(sbin, group):
        """ Import a G1 point from bytes.
        
            Export:
                >>> G = BpGroup()
                >>> g1 = G.gen1()
                >>> buf = g1.export()
                >>> g1p = G1Elem.from_bytes(buf, G)
                >>> g1 == g1p
                True
        """
        pt_bytes = _FFI.new("unsigned char[]", sbin)

        newpt = G1Elem(group)
        _check( _C.G1_ELEM_oct2point(group.bpg, newpt.elem, pt_bytes, len(sbin), _FFI.NULL) )

        return newpt

    def __del__(self):
        if self.elem != None:
            self._C.G1_ELEM_clear_free(self.elem);


class G2Elem(Ops):

    @staticmethod
    def inf(group):
        """ Returns the element at infinity for G2. """
        zero = G2Elem(group)
        _check( _C.G2_ELEM_set_to_infinity(group.bpg, zero.elem) )
        return zero

    def __init__(self, group):
        """ Returns and element of G2. """
        self.group = group
        self.elem = _C.G2_ELEM_new(group.bpg);

    def __del__(self):
        if self.elem != None:
            self._C.G2_ELEM_clear_free(self.elem);

    def __copy__(self):
        """ Copy the G2 point. """
        newpt = G2Elem(self.group)
        _check( _C.G2_ELEM_copy(newpt.elem, self.elem) )
        return newpt

    def add(self, other):
        """ Returns the sum of two points. """
        newpt = G2Elem(self.group)
        _check( _C.G2_ELEM_add(self.group.bpg, newpt.elem, self.elem, other.elem, _FFI.NULL) )
        return newpt   

    def double(self):
        """ Returns the double of the G2 point. """
        newpt = G2Elem(self.group)
        _check( _C.G2_ELEM_dbl(self.group.bpg, newpt.elem, self.elem, _FFI.NULL) )
        return newpt

    def neg(self):
        """ Returns the inverse point. 
        
            Example:
                >>> g2 = BpGroup().gen2()
                >>> g2.add(g2.neg()).isinf()
                True
        """
        newpt = self.__copy__()
        _check( _C.G2_ELEM_invert(self.group.bpg, newpt.elem, _FFI.NULL))
        return newpt

    def eq(self, other):
        """ Returns True if points are equal.
        
            Example:
                >>> G = BpGroup()
                >>> g2 = G.gen2()
                >>> g2.add(g2).eq(g2.double())
                True
                >>> g2.add(g2) == g2.double()
                True
                >>> g2.eq(g2.double())
                False
                >>> g2 != g2.double()
                True
        """
        resp = _C.G2_ELEM_cmp(self.group.bpg, self.elem, other.elem, _FFI.NULL)
        return (int(resp) == 0)

    def isinf(self):
        """ Returns True if the element is infinity."""
        return int(_C.G2_ELEM_is_at_infinity(self.group.bpg, self.elem)) == 1

    @force_Bn(1)
    def mul(self, scalar):
        """ Multiplies the point with a scalar. 
        
            Example:
                >>> g2 = BpGroup().gen2()
                >>> g2.mul(2).eq(g2.double())
                True
        """
        newpt = G2Elem(self.group)
        _check( _C.G2_ELEM_mul(self.group.bpg, newpt.elem, _FFI.NULL, self.elem, scalar.bn, _FFI.NULL) )
        return newpt

    def export(self, form=POINT_CONVERSION_UNCOMPRESSED):
        """ Export a point to a byte representation. """
        size = int(_C.G2_ELEM_point2oct(self.group.bpg, self.elem, form, _FFI.NULL, 0, _FFI.NULL))
        
        out = _FFI.new("unsigned char[]", size)
        _C.G2_ELEM_point2oct(self.group.bpg, self.elem, form, out, size, _FFI.NULL)
        ret = bytes(_FFI.buffer(out)[:size])

        return ret

    @staticmethod
    def from_bytes(sbin, group):
        """ Import a G2 point from bytes.

            Export:
                >>> G = BpGroup()
                >>> g2 = G.gen2()
                >>> buf = g2.export()
                >>> g2p = G2Elem.from_bytes(buf, G)
                >>> g2.eq(g2p)
                True
        """
        pt_bytes = _FFI.new("unsigned char[]", sbin)

        newpt = G2Elem(group)
        _check( _C.G2_ELEM_oct2point(group.bpg, newpt.elem, pt_bytes, len(sbin), _FFI.NULL) )

        return newpt

class GTElem(Ops):

    @staticmethod
    def zero(group):
        """ Returns the element at infinity for GT. """
        zero_pt = GTElem(group)
        _check( _C.GT_ELEM_zero(zero_pt.elem) )
        return zero_pt

    def iszero(self):
        """ Return True if the element is zero."""
        return int(_C.GT_ELEM_is_zero(self.elem)) == 1

    @staticmethod
    def one(group):
        """ Returns the element at infinity for GT. """
        one_pt = GTElem(group)
        _check( _C.GT_ELEM_set_to_unity(group.bpg, one_pt.elem) )
        return one_pt

    def isone(self):
        """ Return zero if the element is one. """
        return int(_C.GT_ELEM_is_unity(self.group.bpg, self.elem)) == 1

    def __init__(self, group):
        """ Returns and element of GT. """
        self.group = group
        self.elem = _C.GT_ELEM_new(group.bpg);

    def __del__(self):
        if self.elem != None:
            self._C.GT_clear_free(self.elem);

    def __copy__(self):
        """ Copy the GT element. 
        
        Example:
            >>> G = BpGroup()
            >>> g1, g2 = G.gen1(), G.gen2()
            >>> gt = G.pair(g1, g2)
            >>> gtp = gt.__copy__()
            >>> gt == gtp
            True
        """
        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_copy(newpt.elem, self.elem) )
        return newpt

    def add(self, other):
        """ Returns the sum of two GT elements. 
        
            Example:
                >>> G = BpGroup()
                >>> zero = GTElem.zero(G)
                >>> x = zero.add(zero)
                >>> x.iszero()
                True
                >>> zero + zero == zero
                True
                >>> one = GTElem.one(G)
                >>> zero + one == one
                True
        """
        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_add(self.group.bpg, newpt.elem, self.elem, other.elem, _FFI.NULL) )
        return newpt

    def sub(self, other):
        """ Returns the difference of two GT elements. 

            Example:
                >>> G = BpGroup()
                >>> zero = GTElem.zero(G)
                >>> one = GTElem.one(G)
                >>> x = one.sub(one)
                >>> x.iszero()
                True
                >>> one - one == zero
                True
        """
        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_sub(self.group.bpg, newpt.elem, self.elem, other.elem, _FFI.NULL) )
        return newpt

    def mul(self, other):
        """ Returns the product of two elements. 

            Example:
                >>> G = BpGroup()
                >>> gt = G.pair(G.gen1(), G.gen2())
                >>> gtinv = gt.inv()
                >>> x = gt.mul(gtinv)
                >>> x.isone()
                True
                >>> gt * gtinv == GTElem.one(G)
                True
        """
        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_mul(self.group.bpg, newpt.elem, self.elem, other.elem, _FFI.NULL) )
        return newpt

    def inv(self):
        """ Returns the inverse element. 

            Example:
                >>> G = BpGroup()
                >>> gt = G.pair(G.gen1(), G.gen2())
                >>> gt2 = gt.mul(gt)
                >>> gtp = gt.sqr()
                >>> gtp.eq(gt2)
                True
        """
        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_inv(self.group.bpg, newpt.elem, self.elem, _FFI.NULL))
        return newpt

    def sqr(self):
        """ Returns the square of an element. """
        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_sqr(self.group.bpg, newpt.elem, self.elem, _FFI.NULL))
        return newpt


    def eq(self, other):
        """ Returns True if elements are equal. """
        resp = _C.GT_ELEM_cmp(self.elem, other.elem)
        return (int(resp) == 0)

    @force_Bn(1)
    def exp(self, scalar):
        """ Exponentiates the element with a scalar. 
        
        Example:
            >>> G = BpGroup()
            >>> g = G.pair(G.gen1(), G.gen2())
            >>> g**0 == g.one(G)
            True
            >>> g**3 * g**5 == g**8
            True
            >>> g**10 * g**(-13) == g**(-3)
            True
            >>> g**2 * g**(-2) == g.one(G)
            True
        """
        
        pt_elem = self
        if scalar < 0:
            pt_elem = self.inv()

        newpt = GTElem(self.group)
        _check( _C.GT_ELEM_exp(self.group.bpg, newpt.elem, pt_elem.elem, scalar.bn, _FFI.NULL) )
        return newpt

    def export(self):
        """ Export a GT element to a byte representation. """
        size = int(_C.GT_ELEM_elem2oct(self.group.bpg, self.elem, _FFI.NULL, 0, _FFI.NULL))
        
        out = _FFI.new("unsigned char[]", size)
        _C.GT_ELEM_elem2oct(self.group.bpg, self.elem, out, size, _FFI.NULL)
        ret = bytes(_FFI.buffer(out)[:size])

        return ret

    @staticmethod
    def from_bytes(sbin, group):
        """ Import a GT element from bytes.

            Export:
                >>> G = BpGroup()
                >>> gt = G.pair(G.gen1(), G.gen2())
                >>> buf = gt.export()
                >>> gtp = GTElem.from_bytes(buf, G)
                >>> gt.eq(gtp)
                True
        """
        pt_bytes = _FFI.new("unsigned char[]", sbin)

        newpt = GTElem(group)
        _check( _C.GT_ELEM_oct2elem(group.bpg, newpt.elem, pt_bytes, len(sbin), _FFI.NULL) )

        return newpt

def test_negatives():
    G = BpGroup()
    gt_v1 = G.pair(-G.gen1(), G.gen2())
    gt_v2 = G.pair(G.gen1(), -G.gen2())
    assert gt_v1 == gt_v2

    assert (-5)*G.gen1() == 5 * (-G.gen1())
    assert (-5)*G.gen1() != 5 * (G.gen1())

    assert G.pair(-10 * G.gen1(), G.gen2()) == G.pair(G.gen1(), G.gen2())**(-10)
    assert G.pair(10 * G.gen1(), -G.gen2()) == G.pair(G.gen1(), G.gen2())**(-10)





import os
import platform
import cffi

# try:
from ._bplib import ffi, lib
_FFI = ffi
_C = lib
#except:
#    print("Support not loading the library to build docs without compiling.")
#    _C = None
#    _FFI = None

def test_load():
	assert _C != None
	assert _FFI != None
	x = _C.BP_GROUP_new()



#!/usr/bin/env python

from setuptools import setup

import bplib

setup(name='bplib_fork',
      version=bplib.VERSION,
      description='A bilinear pairing library for petlib.',
      author='George Danezis',
      author_email='g.danezis@ucl.ac.uk',
      url=r'https://pypi.python.org/pypi/bplib/',
      packages=['bplib'],
      license="LGPL",
      long_description=bplib.__doc__,

      setup_requires=["cffi>=1.0.0",
                      "pytest >= 2.6.4",
                      "petlib_fork >= 0.0.45"],
      package_data={
          "bplib": ["include/*.h"]},
      include_package_data=True,
      tests_require=[
          "cffi >= 1.0.0",
          "pycparser >= 2.10",
          "future >= 0.14.3",
          "pytest >= 2.5.0",
          "pytest-cov >= 1.8.1",
          "petlib_fork >= 0.0.43"
      ],
      cffi_modules=["bplib/compile.py:ffibuilder"],
      install_requires=[
          "cffi >= 1.0.0",
          "pycparser >= 2.10",
          "future >= 0.14.3",
          "pytest >= 2.5.0",
          "pytest-cov >= 1.8.1",
          "msgpack-python >= 0.4.6",
          "petlib_fork >= 0.0.43"
      ],
      zip_safe=False,
      )
